from .defaults import default_filename_template
from .defaults import default_latest_filename_template

import re
from string import Template

from typing import Mapping, Optional, Callable

__all__ = ["generate_info"]

VERSION_REGEX = re.compile(
    r'v?(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(-(?P<status>\w+))?')
SPECIAL_VERSION_NAMES = ["latest", "nightly", "stable"]


rule_sys = {"windows": "winnt", "macos": "mac"}
rules_os = {"windows": "win", "macos": "mac"}
rules_arch = {
    "i686": "x86",
    "x86_64": "x64",
    "ARMv8": "aarch64",
    "ARMv7": "armv7l"
}
rules_osarch = {
    "win-i686": "win32",
    "win-x86_64": "win64",
    "mac-x86_64": "mac64",
    "linux-ARMv7": "linux-armv7l",
    "linux-ARMv8": "linux-aarch64"
}
rules_osbit = {
    "wini686": "win32",
    "winx86_64": "win64",
    "macx86_64": "mac64",
    "linuxARMv7": "linuxarmv7l",
    "linuxARMv8": "linuxaarch64",
    "linuxx86_64": "linux64",
    "linuxi686": "linux32",
    "freebsdx86_64": "freebsd64",
    "freebsdi686": "freebsd32"
}
rules_extension = {
    "windows": "exe",
    "linux": "tar.gz",
    "macos": "dmg",
    "freebsd": "tar.gz"
}
rules_bit = {
    "i686": 32,
    "x86_64": 64,
    "ARMv8": 64,
    "ARMv7": 32,
}

VALID_SYSTEM = ["windows", "linux", "freebsd", "macos"]
VALID_OS = ["win", "linux", "freebsd", "mac"]
VALID_ARCHITECTURE = list(rules_arch.keys())
VALID_ARCH = list(rules_arch.values())


def identity(*args):
    return args[0] if len(args) == 1 else args


def no_validate(*args, **kwargs):
    return True


def is_system(system):
    if system not in VALID_SYSTEM:
        err_msg = f"{system} isn't a valid system, "
        err_msg += f"possible choices are: {', '.join(VALID_SYSTEM)}"
        raise ValueError(err_msg)
    return True


def is_os(os):
    return os in VALID_OS


def is_architecture(arch):
    if arch not in VALID_ARCHITECTURE:
        err_msg = f"{arch} isn't a valid architecture, "
        err_msg += f"possible choices are: {', '.join(VALID_ARCHITECTURE)}"
        raise ValueError(err_msg)
    return True


def is_arch(arch):
    return arch in VALID_ARCH


def is_version(version):
    if version in SPECIAL_VERSION_NAMES:
        return True
    else:
        return bool(VERSION_REGEX.match(version))


def is_valid_release(version, system, architecture):
    if (system == "windows"
            and architecture not in ["i686", "x86_64"]):
        return False
    if (system == "macos"
            and architecture not in ["x86_64"]):
        return False
    if (system == "freebsd"
            and architecture not in ["x86_64"]):
        return False
    if (version == "latest" and (
            architecture not in ["i686", "x86_64"])):
        return False
    return True


class NameFilter:
    def __init__(self,
                 f: Callable = identity,
                 rules: Optional[Mapping] = None,
                 validate: Callable = no_validate):
        self.f = f
        self.rules = rules if rules else {}
        self.validate = validate

    def __call__(self, *args, **kwargs):
        assert self.validate(*args, **kwargs)
        # directly return rst if there're no special filter rules
        rst = self.f(*args, **kwargs)
        return self.rules.get(rst, rst)


f_major_version = NameFilter(lambda x: x.lstrip('v').split('.')[0],
                             validate=is_version)
f_minor_version = NameFilter(lambda x: '.'.join(x.lstrip('v').
                                                split('.')[0:2]),
                             validate=is_version)
f_patch_version = NameFilter(lambda x: x.lstrip('v').split('-')[0],
                             validate=is_version)


def _vmajor_version(ver):
    if ver in SPECIAL_VERSION_NAMES:
        return ver
    else:
        return 'v'+f_major_version(ver)


def _vminor_version(ver):
    if ver in SPECIAL_VERSION_NAMES:
        return ver
    else:
        return 'v'+f_minor_version(ver)


def _vpatch_version(ver):
    if ver in SPECIAL_VERSION_NAMES:
        return ver
    else:
        return 'v'+f_patch_version(ver)


def _version(ver):
    if ver in SPECIAL_VERSION_NAMES:
        return ver
    else:
        return ver.lstrip('v')


f_vmajor_version = NameFilter(_vmajor_version)
f_Vmajor_version = NameFilter(lambda x: f_vmajor_version(x).capitalize())
f_vminor_version = NameFilter(_vminor_version)
f_Vminor_version = NameFilter(lambda x: f_vminor_version(x).capitalize())
f_vpatch_version = NameFilter(_vpatch_version)
f_Vpatch_version = NameFilter(lambda x: f_vpatch_version(x).capitalize())

f_version = NameFilter(_version, validate=is_version)

f_system = NameFilter(validate=is_system)
f_System = NameFilter(f=lambda x: f_system(x).capitalize())
f_SYSTEM = NameFilter(f=lambda x: f_system(x).upper())

f_sys = NameFilter(rules=rule_sys, validate=is_system)
f_Sys = NameFilter(f=lambda x: f_sys(x).capitalize())
f_SYS = NameFilter(f=lambda x: f_sys(x).upper())

f_os = NameFilter(rules=rules_os, validate=is_system)
f_Os = NameFilter(f=lambda x: f_os(x).capitalize())
f_OS = NameFilter(f=lambda x: f_os(x).upper())

f_arch = NameFilter(rules=rules_arch, validate=is_architecture)
f_Arch = NameFilter(f=lambda x: f_arch(x).capitalize())
f_ARCH = NameFilter(f=lambda x: f_arch(x).upper())

f_osarch = NameFilter(f=lambda os, arch: f"{os}-{arch}",
                      rules=rules_osarch,
                      validate=lambda os, arch:
                      is_os(os) and is_architecture(arch))


def _Osarch(os, arch):
    if os in ["win", "mac"]:
        return f_osarch(os, arch).capitalize()
    os, arch = f_osarch(os, arch).split('-')
    return os.capitalize() + '-' + arch


def _OSarch(os, arch):
    if os in ["win", "mac"]:
        return f_osarch(os, arch).upper()
    os, arch = f_osarch(os, arch).split('-')
    return os.upper() + '-' + arch


f_Osarch = NameFilter(_Osarch)
f_OSarch = NameFilter(_OSarch)

f_osbit = NameFilter(f=lambda os, arch: f"{os}{arch}",
                     rules=rules_osbit,
                     validate=lambda os, arch:
                     is_os(os) and is_architecture(arch))

f_bit = NameFilter(rules=rules_bit, validate=is_architecture)

f_extension = NameFilter(rules=rules_extension, validate=is_system)


def _meta_filename(t, *args, **kwargs):
    if not isinstance(t, Template):
        t = Template(t)
    return t.substitute(*args, **kwargs)


def _filename(*args, **kwargs):
    return _meta_filename(default_filename_template, *args, **kwargs)


def _latest_filename(**kwargs):
    return _meta_filename(default_latest_filename_template, **kwargs)


f_filename = NameFilter(_filename)
f_latest_filename = NameFilter(_latest_filename)


def generate_info(plain_version: str,
                  system: str,
                  architecture: str,
                  **kwargs):
    os = f_os(system)
    arch = f_arch(architecture)

    configs = {}
    configs.update(kwargs)

    configs.update({
        "system": system,
        "System": f_System(system),
        "SYSTEM": f_SYSTEM(system),

        "sys": f_sys(system),
        "Sys": f_Sys(system),
        "SYS": f_SYS(system),

        "os": os,
        "Os": f_Os(system),
        "OS": f_OS(system),

        "architecture": architecture,

        "Arch": f_Arch(architecture),
        "arch": arch,
        "ARCH": f_ARCH(architecture),

        "osarch": f_osarch(os, architecture),
        "Osarch": f_Osarch(os, architecture),
        "OSarch": f_OSarch(os, architecture),

        "osbit": f_osbit(os, architecture),

        "bit": f_bit(architecture),
        "extension": f_extension(system),

        "version": f_version(plain_version),
        "major_version": f_major_version(plain_version),
        "vmajor_version": f_vmajor_version(plain_version),
        "minor_version": f_minor_version(plain_version),
        "vminor_version": f_vminor_version(plain_version),
        "patch_version": f_patch_version(plain_version),
        "vpatch_version": f_vpatch_version(plain_version)
    })

    configs.update({
        "filename": f_filename(**configs),
        "latest_filename": f_latest_filename(**configs)
    })

    return configs



from .sys_utils import current_system
import os

PKG_ROOT = os.path.abspath(os.path.dirname(__file__))


def get_configfiles(filename):
    """generate a list of config files ordered with their priorities"""
    configfile_list = []
    sys = current_system()
    if sys in ["macos", "freebsd", "linux"]:
        configfile_list.append(
            os.path.join(os.path.expanduser("~"),
                         ".config", "jill", filename)
        )
    elif sys == "windows":
        configfile_list.append(
            os.path.join(os.path.expanduser(r"~\AppData\Local\julias"),
                         filename)
        )

    # fallback config
    configfile_list.append(os.path.join(PKG_ROOT, "config", filename))
    return configfile_list


SOURCE_CONFIGFILE = get_configfiles("sources.json")
RELEASE_CONFIGFILE = os.path.join(PKG_ROOT, "config", "releases.csv")
GPG_PUBLIC_KEY_PATH = os.path.join(PKG_ROOT, ".gnupg", "juliareleases.asc")

default_filename_template = "julia-$version-$osarch.$extension"
default_latest_filename_template = "julia-latest-$osbit.$extension"

# for mirror usage: where releases are downloaded to
default_path_template = "releases/$vminor_version/$filename"

# ports
default_scheme_ports = {
    "rsync": 873,
    "https": 443,
    "http": 80,
    "ftp": 21,
}



from .source import SourceRegistry
from .filters import is_valid_release
from .filters import SPECIAL_VERSION_NAMES
from .filters import is_system, is_architecture
from .filters import VALID_SYSTEM, VALID_ARCHITECTURE
from .defaults import RELEASE_CONFIGFILE
from .sys_utils import current_system, current_architecture
from semantic_version import Version

from itertools import product
import csv
import os
import logging

from typing import Tuple, List


def read_releases(stable_only=False) -> List[Tuple]:
    cfg_file = RELEASE_CONFIGFILE
    if not os.path.isfile(cfg_file):
        return []
    with open(cfg_file, "r") as csvfile:
        releases = []
        for row in csv.reader(csvfile):
            releases.append(tuple(row))
    if stable_only:
        releases = list(filter(lambda x: x[0] not in SPECIAL_VERSION_NAMES,
                               releases))
    return releases


def is_full_version(version):
    return len(str(version).split("-")[0].split(".")) >= 3


def is_version_released(version, system, architecture,
                        update=False,
                        upstream=None,
                        timeout=5,
                        cache=dict()):
    if not is_valid_release(version, system, architecture):
        return False

    if not cache:
        for item in read_releases():
            cache[item] = True

    item = str(version), system, architecture
    if item in cache:
        return cache[item]

    rst = False
    if update:
        # query process is time-consuming
        registry = SourceRegistry(upstream=upstream)
        rst = bool(registry.query_download_url(*item,
                                               timeout=timeout,
                                               max_try=1))
        if rst:
            logging.info(f"get new release {item}")
            with open(RELEASE_CONFIGFILE, 'a') as csvfile:
                writer = csv.writer(csvfile)
                writer.writerow(item)
        else:
            logging.debug(f"queried {item} = {rst}")
        # only update cache in update mode
        cache[item] = rst
    return rst


def get_version(version: str):
    """
    add tailing 0s for incomplete version string
    """
    if version in ["latest"]:
        return Version("999.999.999")
    splited = str(version).split(".")
    if len(splited) == 1:
        major, minor, patch = splited[0], "0", "0"
    elif len(splited) == 2:
        major, minor = splited[0:2]
        patch = "0"
    elif len(splited) == 3:
        major, minor, patch = splited[0:3]
    else:
        raise ValueError(f"Unrecognized version string {version}")
    return Version(".".join([major, minor, patch]))


def _latest_version(next_version, version, system, architecture,
                    **kwargs) -> str:
    last_version = get_version(version)
    if not is_version_released(last_version, system, architecture,
                               **kwargs):
        return str(last_version)

    while True:
        test_version = next_version(last_version)
        if is_version_released(test_version, system, architecture,
                               **kwargs):
            last_version = test_version
        else:
            return str(last_version)


def latest_patch_version(version, system, architecture, **kwargs) -> str:
    """
    return the latest X.Y.z version starting from input version X.Y.Z
    """
    if not kwargs.get("update", False):
        # just query from the sorted database
        versions = [item for item in read_releases()
                    if (item[2] == architecture and
                        get_version(version).next_minor() > get_version(item[0]))]  # nopep8
        return versions[-1][0]

    return _latest_version(Version.next_patch,
                           str(version), system, architecture,
                           **kwargs)


def latest_minor_version(version, system, architecture, **kwargs) -> str:
    """
    return the latest X.y.z version starting from input version X.Y.Z
    """
    if not kwargs.get("update", False):
        # just query from the sorted database
        versions = [item for item in read_releases()
                    if (item[2] == architecture and
                        get_version(version).next_major() > get_version(item[0]))]  # nopep8
        return versions[-1][0]

    latest_minor = _latest_version(Version.next_minor,
                                   version, system, architecture,
                                   **kwargs)
    latest_patch = latest_patch_version(latest_minor, system, architecture,
                                        **kwargs)
    return latest_patch


def latest_major_version(version, system, architecture, **kwargs) -> str:
    """
    return the latest x.y.z version starting from input version X.Y.Z
    """
    if not kwargs.get("update", False):
        # just query from the sorted database
        versions = [item for item in read_releases()
                    if (item[2] == architecture and
                        get_version("1.0.0").next_major() > get_version(item[0]))]  # nopep8
        return versions[-1][0]

    latest_major = _latest_version(Version.next_major,
                                   version, system, architecture,
                                   **kwargs)
    latest_minor = _latest_version(Version.next_minor,
                                   latest_major, system, architecture,
                                   **kwargs)
    latest_patch = _latest_version(Version.next_patch,
                                   latest_minor, system, architecture,
                                   **kwargs)
    return latest_patch


def latest_version(version, system, architecture, **kwargs) -> str:
    """
    find the latest version for partial semantic version string. Directly
    return `version` if it's already a complete version string.
    """
    if version in SPECIAL_VERSION_NAMES:
        return version
    if is_full_version(version):
        return version

    f_list = [latest_minor_version,
              latest_patch_version]
    f_list.append(lambda ver, sys, arch, **kwargs: ver)  # type: ignore

    if len(version) == 0:
        return latest_major_version('1', system, architecture, **kwargs)
    else:
        idx = len(version.split('.')) - 1
        return f_list[idx](version, system, architecture, **kwargs)


def _make_version(ver: str) -> Version:
    if ver == "latest":
        ver = "999.999.999"
    return Version(ver)


def sort_releases():
    releases = read_releases()
    releases.sort(key=lambda x: (x[1], x[2], _make_version(x[0])))
    with open(RELEASE_CONFIGFILE, 'w') as csvfile:
        for item in releases:
            writer = csv.writer(csvfile)
            writer.writerow(item)


def update_releases(system=None, architecture=None, *,
                    upstream="Official",
                    timeout=5):
    """
    check if there're new Julia releases

    Arguments:
      system: Options are:
        current system(default), SYSTEM, "all"
      architecture: Options are:
        current architecture(default), ARCHITECTURE, "all"
      upstream:
        manually choose a update upstream. By default it only checks
        from JuliaComputing's servers, i.e., "Official" upstream.
      timeout:
        how long each query waits before returning False
    """
    if system == "all":
        systems = VALID_SYSTEM
    else:
        system = system if system else current_system()
        is_system(system)
        systems = [system]

    if architecture == "all":
        architectures = VALID_ARCHITECTURE
    else:
        architecture = architecture if architecture else current_architecture()
        is_architecture(architecture)
        architectures = [architecture]

    # the first run gets all x.y.0 versions
    # the second run gets all x.y.z versions
    logging.info("check new Julia release info, it will take a while")
    kwargs = {
        "update": True,
        "upstream": upstream,
        "timeout": timeout
    }
    for _ in range(2):
        versions = set(map(lambda x: x[0], read_releases(stable_only=True)))
        versions = versions if len(versions) != 0 else {"1.0.0"}

        for item in product(versions, systems, architectures):
            # get all x.y.0 versions
            latest_major_version(*item, **kwargs)
            latest_minor_version(*item, **kwargs)
            latest_patch_version(*item, **kwargs)

    special_versions = ["latest"]
    for item in product(special_versions, systems, architectures):
        is_version_released(*item, **kwargs)

    sort_releases()



from .download import download_package
from .install import install_julia
from .mirror import mirror
from .version_utils import update_releases
from .source import show_upstream
import fire
import logging
import os


def main():
    logging.basicConfig(filename=os.devnull,
                        level=logging.DEBUG)

    logging.basicConfig(level=logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    format = logging.Formatter('%(message)s')
    ch.setFormatter(format)
    logging.getLogger('').addHandler(ch)

    fire.Fire({
        'download': download_package,
        'install': install_julia,
        'mirror': mirror,
        'update': update_releases,
        'upstream': show_upstream
    }, name="jill")


if __name__ == "__main__":
    main()



import platform


def current_system():
    rst = platform.system()
    if rst.lower() == "linux":
        return "linux"
    elif rst.lower() == "darwin":
        return "macos"
    elif rst.lower() == "freebsd":
        return "freebsd"
    elif rst.lower() == "windows":
        return "windows"
    else:
        raise ValueError(f"Unsupported system {rst}")


def current_architecture():
    arch = platform.machine()
    if arch.lower() == "aarch64":
        return "ARMv8"
    elif arch.lower() == "armv7l":
        return "ARMv7"
    elif arch.lower() == "i386":
        return "i686"
    elif arch.lower() == "amd64":
        return "x86_64"
    else:
        return arch



from .defaults import default_scheme_ports
from .defaults import SOURCE_CONFIGFILE

from .net_utils import query_ip, port_response_time
from .net_utils import is_url_available
from .filters import generate_info

from itertools import chain, repeat
from urllib.parse import urlparse
from string import Template

import requests
import json
import os
import logging

from typing import Optional, List

from requests.exceptions import RequestException


class ReleaseSource:
    def __init__(self,
                 name: str,
                 urls: List[str],
                 latest_urls: List[str],
                 timeout=2.0):
        # seperate stable and nightly versions because:
        #   * JuliaComputing stores them in two different s3 buckets
        #   * not all mirror servers need/want to serve nightly releases
        # store multiple urls because:
        #   * many mirrors have different multiple domains for different
        #     network choices, e.g., ipv4, ipv6, cernet, chinanet, rsync
        self.name = name
        self.url_templates = [Template(x) for x in urls]
        # TODO: make latest an optional config
        self.latest_url_templates = [Template(x) for x in latest_urls]
        self.timeout = timeout
        self._latencies = dict()  # type: ignore

    @property
    def urls(self):
        return [x.template for x in self.url_templates]

    @property
    def latest_urls(self):
        return [x.template for x in self.latest_url_templates]

    @property
    def hosts(self):
        return [urlparse(url).netloc for url in self.urls]

    @property
    def latest_hosts(self):
        return [urlparse(url).netloc for url in self.latest_urls]

    @property
    def latencies(self):
        # only check latency once and lazily
        if not self._latencies:
            url_list = self.urls.copy()
            url_list.extend(self.latest_urls)
            host_list = [urlparse(url).netloc for url in url_list]

            # hosts might point to the same ip address
            host_ip_records = {host: query_ip(host) for host in host_list}
            ip_port_records = {host_ip_records[urlparse(url).netloc]:
                               default_scheme_ports[urlparse(url).scheme]
                               for url in url_list}

            latency_records = dict()
            # TODO: use threads
            for ip, port in ip_port_records.items():
                latency_records[ip] = port_response_time(ip, port,
                                                         self.timeout)
            self._latencies = {host: latency_records[host_ip_records[host]]
                               for host in host_list}
        return self._latencies

    def __repr__(self):
        return f"ReleaseSource('{self.name}')"

    def get_url(self, plain_version, system, architecture):
        """
        return one potential downloading url with minal network latency for
        specific version, system and architecture. Special version name such
        as 'latest' are treated differently.
        """
        if plain_version == "latest":
            template_lists = self.latest_url_templates
        else:
            template_lists = self.url_templates
        configs = generate_info(plain_version, system, architecture)
        url_list = [t.substitute(**configs) for t in template_lists]
        url_list.sort(key=lambda url:
                      self.latencies[urlparse(url).netloc])
        return url_list[0]


def read_registry():
    registry = dict()
    for cfg_file in reversed(SOURCE_CONFIGFILE):
        if not os.path.isfile(cfg_file):
            continue
        with open(cfg_file, 'r') as f:
            upstream_records = json.load(f).get("upstream", {})
            temp_registry = {k: ReleaseSource(**v)
                             for k, v in upstream_records.items()}
        registry.update(temp_registry)
    return registry


class SourceRegistry:
    # share the same "full" registry across all instances
    class __SourceRegistry:
        def __init__(self, timeout):
            self.timeout = timeout
            self.registry = read_registry()

    __inner_registry = None

    def __init__(self, *, upstream=None, timeout=2):
        self.upstream = upstream
        if not self.__inner_registry:
            self.__inner_registry = self.__SourceRegistry(timeout)

    @property
    def registry(self):
        if self.upstream:
            # users limit themselves to only one download source
            if self.upstream in self.__inner_registry.registry:
                return {self.upstream:
                        self.__inner_registry.registry[self.upstream]}
            else:
                msg = "valid sources are:" + \
                    ', '.join(self.__inner_registry.registry.keys())
                raise ValueError(msg)
        return self.__inner_registry.registry

    @property
    def latencies(self):
        records = dict()
        for src in self.registry.values():
            records.update(src.latencies)
        return records

    def __len__(self):
        return len(self.registry)

    def info(self):
        msg = f"found {len(self)} release sources:\n"
        for src in self.registry.keys():
            msg += "  - " + str(src) + "\n"
        return msg

    def _get_urls(self, plain_version, system, architecture):
        """
        return a list of potential downloading urls for specific version,
        system and architecture. Special version name such as 'latest' are
        treated differently.
        """
        url_list = [src.get_url(plain_version, system, architecture)
                    for src in self.registry.values()]
        url_list.sort(key=lambda url:
                      self.latencies[urlparse(url).netloc])
        return url_list

    def query_download_url(self,
                           version, system, arch, *,
                           max_try=3, timeout=10):
        """
        return a valid download url to nearest mirror server. If there isn't
        such version then return None.
        """
        url_list = self._get_urls(version, system, arch)

        url_list = chain.from_iterable(repeat(url_list, max_try))
        for url in url_list:
            if is_url_available(url, timeout):
                return url
        return None


def show_upstream():
    """print all registered upstream servers"""
    registry = SourceRegistry()
    print(registry.info())



from .defaults import GPG_PUBLIC_KEY_PATH

from gnupg import GPG
from tempfile import TemporaryDirectory

import os


def _verify_gpg(gpg: GPG, datafile, signature_file):
    with open(signature_file, "rb") as fh:
        return gpg.verify_file(fh, data_filename=datafile)


def verify_gpg(datafile, signature_file=None) -> bool:
    """
    verify Julia releases using GPG
    """
    if signature_file is None:
        signature_file = datafile + ".asc"

    with open(GPG_PUBLIC_KEY_PATH) as fh:
        keycontent = fh.read()
    with TemporaryDirectory() as tmpdir:
        gpg = GPG(gnupghome=tmpdir)
        gpg.import_keys(keycontent)
        return bool(_verify_gpg(gpg, datafile, signature_file))



from .source import SourceRegistry
from .version_utils import latest_version
from .version_utils import is_version_released
from .version_utils import is_full_version
from .sys_utils import current_system, current_architecture
from .gpg_utils import verify_gpg

import wget
import os
import shutil
import tempfile
import logging

from urllib.parse import urlparse

from typing import Optional

from urllib.error import URLError


def _download(url: str, out: str):
    # always do overwrite
    outpath = os.path.abspath(out)
    outdir, outname = os.path.split(outpath)

    with tempfile.TemporaryDirectory() as temp_outdir:
        temp_outpath = os.path.join(temp_outdir, outname)
        try:
            logging.info(f"downloading source: {url}")
            wget.download(url, temp_outpath)
            print()  # for format usage
            logging.info(f"finished downloading {outname}")
        except (URLError, ConnectionError) as e:
            logging.warning(f"failed to download {outname}")
            return False

        if not os.path.isdir(outdir):
            os.makedirs(outdir, exist_ok=True)
        shutil.move(temp_outpath, outpath)

    return outpath


def download_package(version=None, sys=None, arch=None, *,
                     upstream=None,
                     outdir=None,
                     overwrite=False,
                     max_try=3):
    """
    download julia release from nearest servers

    Arguments:
      version: Option examples: 1, 1.2, 1.2.3, latest.
      sys: Options are: "linux", "macos", "freebsd", "windows"
      arch: Options are: "i686", "x86_64", "ARMv7", "ARMv8"
      upstream:
        manually choose a download upstream. For example, set it to "Official"
        if you want to download from JuliaComputing's s3 buckets.
      outdir: where release is downloaded to. By default it's current folder.
      overwrite: True to overwrite existing releases. By default it's False.
      max_try: try `max_try` times before returning a False.
    """
    version = str(version) if version else ''
    version = "latest" if version == "nightly" else version
    version = "" if version == "stable" else version

    system = sys if sys else current_system()
    architecture = arch if arch else current_architecture()

    # allow downloading unregistered releases, e.g., 1.4.0-rc1
    do_release_check = not is_full_version(version)
    version = latest_version(version, system, architecture)

    if architecture in ["ARMv7", "ARMv8"]:
        # TODO: fix update functionality for it in version_utils
        msg = f"update is disabled for tier-2 support {architecture}"
        logging.warning(msg)
        update = False
    else:
        update = True

    release_str = f"{version}-{system}-{architecture}"
    if do_release_check:
        rst = is_version_released(version, system, architecture,
                                  update=update)
        if not rst:
            msg = f"failed to find Julia release for {release_str}."
            logging.info(msg)
            return False

    logging.info(f"download Julia release for {release_str}")
    registry = SourceRegistry(upstream=upstream)
    url = registry.query_download_url(version, system, architecture,
                                      max_try=max_try)
    if not url:
        msg = f"failed to find available upstream for {release_str}"
        logging.warning(msg)
        return None

    outdir = outdir if outdir else '.'
    outdir = os.path.abspath(outdir)
    if current_system() == "windows":
        outdir = outdir.replace("\\\\", "\\")

    outname = os.path.split(urlparse(url).path)[1]
    outpath = os.path.join(outdir, outname)

    skip_download = False
    if system in ["linux", "freebsd"]:
        if (os.path.isfile(outpath) and os.path.isfile(outpath+".asc")
                and not overwrite):
            skip_download = True
    else:
        if os.path.isfile(outpath) and not overwrite:
            skip_download = True
    if skip_download:
        logging.info(f"{outname} already exists, skip downloading")
        return outpath

    package_path = _download(url, outpath)

    if system in ["windows", "macos"]:
        # macOS and Windows releases are codesigned with certificates
        # that are verified by the operating system during installation
        return package_path
    elif system in ["linux", "freebsd"]:
        # additional verification using GPG
        if not package_path:
            return package_path

        # a mirror should provides both *.tar.gz and *.tar.gz.asc
        gpg_signature_path = _download(url+".asc", outpath+".asc")
        if not gpg_signature_path:
            msg = f"failed to download GPG signature for {release_str}"
            logging.warning(msg)
            logging.info(f"remove untrusted/broken file")
            os.remove(package_path)
            return False

        if not verify_gpg(package_path, gpg_signature_path):
            logging.warning(f"failed to verify {release_str} downloads")
            msg = f"remove untrusted/broken files"
            logging.info(msg)
            os.remove(package_path)
            os.remove(gpg_signature_path)
            return False

        # GPG-verified julia release path
        logging.info(f"success to verify {release_str} downloads")
        return package_path
    else:
        raise ValueError(f"unsupported system {sys}")



from .filters import f_major_version, f_minor_version, f_patch_version
from .filters import SPECIAL_VERSION_NAMES
from .download import download_package
from .interactive_utils import query_yes_no
from .sys_utils import current_architecture, current_system
from .version_utils import latest_version
from .mount_utils import DmgMounter, TarMounter
from semantic_version import Version
import os
import getpass
import shutil
import subprocess
import logging


def default_depot_path():
    return os.path.expanduser("~/.julia")


def default_symlink_dir():
    system = current_system()
    if system == "windows":
        return os.path.expanduser(r"~\AppData\Local\julias\bin")
    if getpass.getuser() == "root":
        # available to all users
        return "/usr/local/bin"
    else:
        # exclusive to current user
        return os.path.expanduser("~/.local/bin")


def default_install_dir():
    system = current_system()
    if system == "macos":
        return "/Applications"
    elif system in ["linux", "freebsd"]:
        if getpass.getuser() == "root":
            return "/opt/julias"
        else:
            return os.path.expanduser("~/packages/julias")
    elif system == "windows":
        return os.path.expanduser(r"~\AppData\Local\julias")
    else:
        raise ValueError(f"Unsupported system: {system}")


def get_exec_version(path):
    ver_cmd = [path, "--version"]
    try:
        # outputs: "julia version 1.4.0-rc1"
        version = subprocess.check_output(ver_cmd).decode("utf-8")
        version = version.lower().split("version")[-1].strip()
    except subprocess.CalledProcessError as e:
        # in case it's not executable
        logging.warn(e)
        version = "0.0.1"
    return version


def last_julia_version(version=None):
    # version should follow semantic version syntax
    def sort_key(ver):
        return float(ver.lstrip("v"))

    version = float(f_minor_version(version)) if version else 999.999
    proj_versions = os.listdir(os.path.join(default_depot_path(),
                                            "environments"))
    proj_versions = sorted(filter(lambda ver: sort_key(ver) < version,
                                  proj_versions),
                           key=sort_key)
    if proj_versions:
        return proj_versions[-1]
    else:
        return None


def make_symlinks(src_bin, symlink_dir, version):
    if not os.path.isfile(src_bin):
        raise(ValueError(f"{src_bin} doesn't exist."))

    system = current_system()
    if symlink_dir not in os.environ["PATH"].split(os.pathsep):
        logging.info(f"add {symlink_dir} to PATH")
        if system == "windows":
            # FIXME: this alse copies system PATH to user PATH
            subprocess.run(["powershell.exe",
                            "setx", "PATH", f'"$env:PATH;{symlink_dir}"'])
        else:
            msg = "~/.bashrc will be modified"
            msg += ", if you're not using BASH, then you'll need manually"
            msg += f" add {symlink_dir} to your PATH\n"
            logging.info(msg)

            rc_file = os.path.expanduser("~/.bashrc")
            logging.info(f"{rc_file} is modified")
            with open(rc_file, "a") as file:
                file.writelines("\n# added by jill\n")
                file.writelines(f"export PATH={symlink_dir}:$PATH\n")
        logging.info(f"you need to restart your current shell to update PATH")

    os.makedirs(symlink_dir, exist_ok=True)

    if version in SPECIAL_VERSION_NAMES:
        # issue 11: don't symlink to julia
        link_list = [f"julia-{f_major_version(version)}"]
    else:
        link_list = [f"julia-{f(version)}" for f in (f_major_version,
                                                     f_minor_version,
                                                     f_patch_version)]
        link_list.append("julia")

    for linkname in link_list:
        linkpath = os.path.join(symlink_dir, linkname)
        if current_system() == "windows":
            linkpath += ".cmd"
        if os.path.exists(linkpath) or os.path.islink(linkpath):
            old_ver = get_exec_version(linkpath)
            if version != "latest" and Version(old_ver) > Version(version):
                continue
            logging.info(f"removing previous symlink {linkname}")
            os.remove(linkpath)
        logging.info(f"make symlink {linkpath}")
        if current_system() == "windows":
            with open(linkpath, 'w') as f:
                # create a cmd file to mimic how we do symlinks in linux
                f.writelines(["@echo off\n", f"{src_bin} %*"])
        else:
            os.symlink(src_bin, linkpath)


def copy_root_project(version):
    mver = f_minor_version(version)
    old_ver = last_julia_version(version)
    if old_ver is None:
        logging.info(
            f"Can't find available old root project for version f{version}")
        return None

    env_path = os.path.join(default_depot_path(), "environments")
    src_path = os.path.join(env_path, old_ver)
    dest_path = os.path.join(env_path, f"v{mver}")

    if os.path.exists(dest_path):
        bak_path = os.path.join(env_path, f"v{mver}.bak")
        logging.info(f"move {dest_path} to {bak_path}")
        shutil.move(dest_path, bak_path)
    shutil.copytree(src_path, dest_path)


def install_julia_linux(package_path,
                        install_dir,
                        symlink_dir,
                        version,
                        upgrade):
    mver = f_minor_version(version)
    with TarMounter(package_path) as root:
        src_path = root
        dest_path = os.path.join(install_dir, f"julia-{mver}")
        if os.path.exists(dest_path):
            logging.info(f"remove previous julia installation: {dest_path}")
            shutil.rmtree(dest_path)
        shutil.copytree(src_path, dest_path)
    os.chmod(dest_path, 0o755)  # issue 12
    bin_path = os.path.join(dest_path, "bin", "julia")
    make_symlinks(bin_path, symlink_dir, version)
    if upgrade:
        copy_root_project(version)
    return True


def install_julia_mac(package_path,
                      install_dir,
                      symlink_dir,
                      version,
                      upgrade):
    assert os.path.splitext(package_path)[1] == ".dmg"
    with DmgMounter(package_path) as root:
        # mounted image contents:
        #   ['.VolumeIcon.icns', 'Applications', 'Julia-1.3.app']
        appname = next(filter(lambda x: x.lower().startswith('julia'),
                              os.listdir(root)))
        src_path = os.path.join(root, appname)
        dest_path = os.path.join(install_dir, appname)
        if os.path.exists(dest_path):
            logging.info(f"remove previous julia installation: {dest_path}")
            shutil.rmtree(dest_path)
        shutil.copytree(src_path, dest_path)
    bin_path = os.path.join(dest_path,
                            "Contents", "Resources", "julia", "bin", "julia")
    make_symlinks(bin_path, symlink_dir, version)
    if upgrade:
        copy_root_project(version)
    return True


def install_julia_windows(package_path,
                          install_dir,
                          symlink_dir,
                          version,
                          upgrade):
    assert os.path.splitext(package_path)[1] == ".exe"

    dest_path = os.path.join(install_dir,
                             f"julia-{f_minor_version(version)}")
    if os.path.exists(dest_path):
        logging.info(f"remove previous julia installation: {dest_path}")
        shutil.rmtree(dest_path)

    if Version(version).next_patch() < Version("1.4.0"):
        subprocess.check_output([f'{package_path}',
                                 '/S', f'/D={dest_path}'])
    else:
        subprocess.check_output([f'{package_path}',
                                 '/VERYSILENT',
                                 f'/DIR={dest_path}'])
    bin_path = os.path.join(dest_path, "bin", "julia.exe")
    make_symlinks(bin_path, symlink_dir, version)
    if upgrade:
        copy_root_project(version)
    return True


def install_julia(version=None, *,
                  install_dir=None,
                  symlink_dir=None,
                  upgrade=False,
                  upstream=None,
                  keep_downloads=False,
                  confirm=False):
    """
    Install julia for Linux and MacOS

    Arguments:
      version: Option examples: 1, 1.2, 1.2.3, latest.
      upstream:
        manually choose a download upstream. For example, set it to "Official"
        if you want to download from JuliaComputing's s3 buckets.
      upgrade: True to also copy the root environment from older julia version.
      keep_downloads: True to not remove downloaded releases.
      confirm: add `--confirm` to skip interactive prompt.
    """
    install_dir = install_dir if install_dir else default_install_dir()
    symlink_dir = symlink_dir if symlink_dir else default_symlink_dir()
    system, arch = current_system(), current_architecture()
    version = str(version) if version else ''
    version = "latest" if version == "nightly" else version
    version = "" if version == "stable" else version

    if system == "windows":
        install_dir = install_dir.replace("\\\\", "\\")
    if not confirm:
        version_str = version if version else "latest release"
        question = "jill will:\n"
        question += f"    1) install Julia {version_str} for {system}-{arch}"
        question += f" into {install_dir}\n"
        question += f"    2) make symlinks in {symlink_dir}\n"
        question += f"    3) add {symlink_dir} to PATH if necessary\n"
        question += "Continue installation?"
        to_continue = query_yes_no(question)
        if not to_continue:
            return False

    overwrite = True if version == "latest" else False
    package_path = download_package(version, system, arch,
                                    upstream=upstream,
                                    overwrite=overwrite)
    version = latest_version(version, system, arch)
    if not package_path:
        return False

    if system == "macos":
        installer = install_julia_mac
    elif system in ["linux", "freebsd"]:
        installer = install_julia_linux
    elif system == "windows":
        installer = install_julia_windows
    else:
        raise ValueError(f"Unsupported system: {system}")

    installer(package_path, install_dir, symlink_dir, version, upgrade)

    if not keep_downloads:
        logging.info("remove downloaded files")
        logging.info(f"remove {package_path}")
        os.remove(package_path)
        gpg_signature_file = package_path + ".asc"
        if os.path.exists(gpg_signature_file):
            logging.info(f"remove {gpg_signature_file}")
            os.remove(gpg_signature_file)
    return True



from setuptools import find_packages

import setuptools
import os
from io import open as io_open

src_dir = os.path.abspath(os.path.dirname(__file__))

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = os.path.join(src_dir, 'requirements.txt')
with io_open(requirements, mode='r') as fd:
    install_requires = [i.strip().split('#', 1)[0].strip()
                        for i in fd.read().strip().split('\n')]

setuptools.setup(
    name='jill',
    version='0.5.0',
    author="Johnny Chen",
    author_email="johnnychen94@hotmail.com",
    description="Julia Installer 4 Linux(and MacOS) - Light",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/johnnychen94/jill.py",
    packages=['jill'] + ['jill.' + i for i in find_packages('jill')],
    provides=['jill'],
    install_requires=install_requires,
    python_requires=">=3.6",
    entry_points={'console_scripts': ['jill=jill.__main__:main'], },
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
